/*
 *  guid_utils.c
 *  Implements native procedures for converting guids from bitstrings to hexstrings
 *
 *  Copyright 2007 Unity Technologies Aps. All rights reserved.
 *
 */

#include "postgres.h"
#include "fmgr.h"
#include "utils/varbit.h"


inline char nibble2hex (int n);
inline int hex2nibble (char c);

Datum guid2hex(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(guid2hex);
Datum hex2guid(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(hex2guid);


inline char nibble2hex (int n) {
	int masked=n&15;
	return (masked < 10)?(masked + '0'):(masked - 10 + 'a'); 
}

inline int hex2nibble (char c) {
	if(c >= '0' && c <= '9')
		return c-'0';
	if(c >= 'a' && c <= 'g')
		return c-'a'+10;
	if(c >= 'A' && c <= 'G')
		return c-'A'+10;
	return 0;
}


Datum guid2hex(PG_FUNCTION_ARGS)
{
	VarBit*	 bit;
	VarChar* result;
	int size; 
	int textSize;
	int i;
	
	if(PG_ARGISNULL(0)) PG_RETURN_NULL();
	
	bit = PG_GETARG_VARBIT_P(0);
	size = VARBITBYTES(bit);
	textSize = size * 2;
	
	result = palloc(textSize + VARHDRSZ);
	VARATT_SIZEP(result) = textSize + VARHDRSZ;
	for(i=0; i < size; i++) {
		((char*)VARDATA(result))[i*2]=nibble2hex(bit->bit_dat[i]>>4);
		((char*)VARDATA(result))[i*2+1]=nibble2hex(bit->bit_dat[i]);
	}
	PG_RETURN_VARCHAR_P(result);
}

Datum hex2guid(PG_FUNCTION_ARGS)
{
	VarChar* hex;
	VarBit* result;
	char* data;
	int size; 
	int textSize;
	int i, acc;
	
	if(PG_ARGISNULL(0)) PG_RETURN_NULL();
	
	hex = PG_GETARG_VARCHAR_P(0);
	textSize = VARATT_SIZEP(hex)-VARHDRSZ;
	data = VARDATA(hex);
	size = textSize/2;
	
	
	result = palloc(size + VARHDRSZ + VARBITHDRSZ);
	VARATT_SIZEP(result) = size + VARHDRSZ + VARBITHDRSZ;
	result->bit_len=size*8;
	
	acc=0;
	for(i=0; i < size; i++) {
		result->bit_dat[i] = hex2nibble(data[i*2]) << 4 | hex2nibble(data[i*2+1]);
		acc |= result->bit_dat[i];
	}
	
	// Special case: all zeroes result in a NULL result
	if(acc == 0) {
		pfree(result);
		PG_RETURN_NULL();
	}
	PG_RETURN_VARBIT_P(result);
}

