/*------------------------------------------------------------------------- *
 * nodeConn.c
 *	  Routines to handle connecting of relations.
 *
 * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
 * Portions Copyright (c) 1994, Regents of the University of California
 *
 * Based on nodeSort.c, by Evgen Potemkin evgent@terminal.ru, 11.2002
 *
 *-------------------------------------------------------------------------
 */

#include "postgres.h"

#include "executor/execdebug.h"
#include "executor/nodeConn.h"
#include "utils/tupleconn.h"
#include "miscadmin.h"
#include "utils/memutils.h"

#include "nodes/print.h"

/* ----------------------------------------------------------------
 *		ExecConn
 *
 *		Connects tuples from the outer subtree of the node using tupleconn,
 *		which saves the results in a temporary file or memory. After the
 *		initial call, returns a tuple from the file with each call.
 *
 *		Conditions:
 *		  -- none.
 *
 *		Initial States:
 *		  -- the outer child is prepared to return the first tuple.
 * ----------------------------------------------------------------
 */
TupleTableSlot *
ExecConn(ConnectState *node)
{
	EState	   *estate;
	TupleTableSlot *slot;
	HeapTuple	heapTuple;
	ScanDirection dir;
	Tupleconnstate *tupleconnstate;
	bool		should_free;
	List		*squal;
	ExprContext *econtext;
	int			head;
	bool		ok;
  Conn *plan=(Conn *) node->ss.ps.plan;
	/*
	 * get state info from node
	 */
	estate = node->ss.ps.state;
	dir = estate->es_direction;
	tupleconnstate = (Tupleconnstate *) node->tupleconnstate;
	
  econtext = node->ss.ps.ps_ExprContext;
	if (!node->conn_Done)
	{
		TupleDesc	tupDesc;
    PlanState *outerNode;
    
  	SO1_printf("ExecConn: %s\n",
				   "connecting subplan");

		/*
		 * Want to scan subplan in the forward direction while creating
		 * the data for processing.
		 */

		/*
		 * Initialize tupleconn module.
		 */
		SO1_printf("ExecConn: %s\n",
				   "calling tupleconn_begin");

		outerNode = outerPlanState((PlanState *) node);
		tupDesc = ExecGetResultType(outerNode);

    tupleconnstate = tupleconn_begin_heap(work_mem,tupDesc,plan->connOp,
         plan->priorsList,plan->levelColNum,((Plan *)plan)->plan_rows,plan->useHash);
		node->tupleconnstate = (void *) tupleconnstate;

		/*
		 * Scan the subplan and feed all the tuples to tupleconn.
		 */
    squal = list_make1(node->startQual);

		ResetExprContext(econtext);

		for (;;)
		{
			slot = ExecProcNode(outerNode);
			if (TupIsNull(slot))
				break;
			econtext->ecxt_scantuple = slot;
			/* check if tuple is a head of tree */
			head = ExecQual(squal, econtext, false);
			/* store it */
			tupleconn_puttuple(tupleconnstate, (void *) ExecFetchSlotTuple(slot),head);
		}

		tupleconn_donestoring(tupleconnstate);
		/* connect them.
		 * parentExpr, childExpr, econtext is not needed in tupleconn,
		 * so pass them only for connecting stage.
		 */
		tupleconn_performconn(tupleconnstate,node->parentExpr,node->childExpr,econtext);
		/*
		 * finally set the connected flag to true
		 */
		node->conn_Done = true;
		SO1_printf(stderr, "ExecConn: connecting done.\n");
	}
	SO1_printf("ExecConn: %s\n",
			   "retrieving tuple from tupleconn");

	/*
	 * Get the first or next tuple from tupleconn. Returns NULL if no more
	 * tuples.
	 */
  squal = node->ss.ps.qual;
	ok=0;	
	/* everywhere here slot is the same pointer, so get it here */
	slot = node->ss.ps.ps_ResultTupleSlot;
	/* Apply HAVING clause to post-qual tuples */
	do{
		/* if qual present, clear context */
		if(squal) ResetExprContext(econtext);
		heapTuple = tupleconn_getheaptuple(tupleconnstate,
									   &should_free);
		if(!heapTuple) {/* no more tuples */
                  ExecClearTuple(slot);
                  break;
                }
		ExecStoreTuple(heapTuple, slot, InvalidBuffer, should_free);
                if(!squal) break; /* don't qual */

		econtext->ecxt_scantuple = slot;
		/* if tuple satisfy qual, pass it, else get next */
		ok=ExecQual(squal, econtext, false);
	}while(!ok);

	if(squal) ResetExprContext(econtext);
	return slot;
}

/* Everything below is same as nodeSort.c
 */
/* ----------------------------------------------------------------
 *		ExecInitConn
 *
 *		Creates the run-time state information for the conn node
 *		produced by the planner and initailizes its outer subtree.
 * ----------------------------------------------------------------
 */
ConnectState *
ExecInitConn(Conn *node, EState *estate, int eflags)
{
	ConnectState *connstate;
	
	SO1_printf("ExecInitConn: %s\n",
			   "initializing conn node");
	
	/* check for unsupported flags */
//	Assert(!(eflags & EXEC_FLAG_MARK));
	
	/*
	 * assign the node's execution state
	 */

	/*
	 * create state structure
	 */
	connstate = makeNode(ConnectState);
	connstate->ss.ps.plan = (Plan *) node;
	connstate->ss.ps.qual = ExecInitExpr( (Expr *)((Plan *)node)->qual, (PlanState *)connstate);;
	connstate->ss.ps.state = estate;

	connstate->startQual = ExecInitExpr( (Expr *)node->startQual, (PlanState *)connstate);
	connstate->parentExpr = ExecInitExpr((Expr *)node->parentExpr,(PlanState *)connstate);
	connstate->childExpr = ExecInitExpr( (Expr *)node->childExpr, (PlanState *)connstate);

  connstate->conn_Done = false;
	connstate->tupleconnstate = NULL;

	/*
	 * Miscellaneous initialization
	 */
  ExecAssignExprContext(estate, &connstate->ss.ps); 
#define CONN_NSLOTS 2

	/*
	 * tuple table initialization
	 */
	ExecInitResultTupleSlot(estate, &connstate->ss.ps);
	ExecInitScanTupleSlot(estate, &connstate->ss);
	/*
	 * initializes child nodes
	 */
        outerPlanState(connstate) = ExecInitNode(outerPlan(node), estate, eflags);
	/*
	 * initialize tuple type.  no need to initialize projection info
	 * because this node doesn't do projections.
	 */
//	ExecAssignResultTypeFromOuterPlan(&connstate->ss.ps);

        ExecAssignResultType(&connstate->ss.ps, ExecGetResultType(outerPlanState(&connstate->ss.ps)));
	ExecAssignScanTypeFromOuterPlan(&connstate->ss);
	connstate->ss.ps.ps_ProjInfo = NULL;

	SO1_printf("ExecInitConn: %s\n",
			   "conn node initialized");

	return connstate;
}

int
ExecCountSlotsConn(Conn *node)
{
	return ExecCountSlotsNode(outerPlan((Plan *) node)) +
		ExecCountSlotsNode(innerPlan((Plan *) node))+CONN_NSLOTS;
}

/* ----------------------------------------------------------------
 *		ExecEndConn(node)
 * ----------------------------------------------------------------
 */
void
ExecEndConn(ConnectState *node)
{
	/*
	 * get info from the conn state
	 */
	SO1_printf("ExecEndConn: %s\n",
			   "shutting down conn node");

	/*
	 * shut down the subplan
	 */

	/*
	 * clean out the tuple table
	 */
	ExecClearTuple(node->ss.ps.ps_ResultTupleSlot);
	ExecClearTuple(node->ss.ss_ScanTupleSlot);

	/*
	 * Release tupleconn resources
	 */
	if (node->tupleconnstate != NULL)
		tupleconn_end((Tupleconnstate *) node->tupleconnstate);
	node->tupleconnstate = NULL;

	ExecEndNode(outerPlanState(node));

	SO1_printf("ExecEndConn: %s\n",
			   "conn node shutdown");
}

/* ----------------------------------------------------------------
 *		ExecConnMarkPos
 *
 *		Calls tupleconn to save the current position in the processed file.
 * ----------------------------------------------------------------
 */
void
ExecConnMarkPos(ConnectState *node)
{

	/*
	 * if we haven't processed yet, just return
	 */
	if (!node->conn_Done)
		return;

	tupleconn_markpos((Tupleconnstate *) node->tupleconnstate);
}

/* ----------------------------------------------------------------
 *		ExecConnRestrPos
 *
 *		Calls tupleconn to restore the last saved processed file position.
 * ----------------------------------------------------------------
 */
void
ExecConnRestrPos(ConnectState *node)
{

	/*
	 * if we haven't processed yet, just return.
	 */
	if (!node->conn_Done)
		return;

	/*
	 * restore the scan to the previously marked position
	 */
	tupleconn_restorepos((Tupleconnstate *) node->tupleconnstate);
}

void
ExecReScanConn(ConnectState *node, ExprContext *exprCtxt)
{

	/*
	 * If we haven't processed yet, just return. If outerplan' chgParam is
	 * not NULL then it will be re-scanned by ExecProcNode, else - no
	 * reason to re-scan it at all.
	 */
	if (!node->conn_Done)
		return;

	ExecClearTuple(node->ss.ps.ps_ResultTupleSlot);

	/*
	 * If subnode is to be rescanned then we forget previous process results;
	 * we have to re-read the subplan and re-connect.
	 *
	 * Otherwise we can just rewind and rescan the processed output.
	 */

	if (((PlanState *) node)->lefttree->chgParam != NULL)
	{
		node->conn_Done = false;
		tupleconn_end((Tupleconnstate *) node->tupleconnstate);
		node->tupleconnstate = NULL;
	}
	else
		tupleconn_rescan((Tupleconnstate *) node->tupleconnstate);

}
